/* session.cc
 * Copyright (C) 2001-2 The University of Newcastle upon Tyne
 *
 * Written by Paul Coates (Paul.Coates@ncl.ac.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <gnome.h>
#include <gtk/gtk.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include "filter.h"
#include "filteritem.h"
#include "session.h"
#include "topology.h"
#include "sessionlist.h"
#include "scheduler.h"
#include "data.h"
#include "main.h"
#include "gui.h"
#include "misc.h"
#include "preferences.h"
#include "monitor.h"
#include "viewsidebar.h"
#include "diagram.h"
#include "view.h"
#include "header.h"

static void close_cb(GtkWidget *w, gpointer data)
{	gtk_widget_hide(GTK_WIDGET(data)); }
static void action_cb(GtkWidget *w, gpointer data)
{ ((Session*)data)->session_action(); }
static void have_title_cb(GtkWidget *w, gpointer data)
{
	((Session*)data)->save_msg.hide();
	((Session*)data)->have_title();
}
static void unique_name_cb(GtkWidget *w, gpointer data)
{
	((Session*)data)->save_session_list.hide();
	((Session*)data)->unique_name();
}
static void save_session_cb(GtkWidget *w, gpointer data)
{
	Session* s = (Session*) data;
	strcpy(s->session_name, s->save_session_list.getName());
	s->save_session();
}
static void load_session_cb(GtkWidget *w, gpointer data)
{
	Session* s = (Session*) data;
	s->load_session(s->load_session_list.getName());
}

Session::Session(Main *m)
{
	struct stat statbuf;
	struct passwd *pwd;
	char buf[256];
	char *md;
	View *v;

	md = getenv("NOMAD_DIR");
	if (md!=NULL) { strcpy(datadir, md); } else
	{ pwd = getpwuid(getuid());
		sprintf(datadir, "%s/%s", pwd->pw_dir, DATA_DIR);
	}

	sprintf(buf, "%s", datadir);
	if (stat(buf, &statbuf) == -1) mkdir(buf, 488); // rwxr x
	sprintf(buf, "%s/sessions", datadir);
	if (stat(buf, &statbuf) == -1) mkdir(buf, 488); // rwxr x

	main = m;
	sprintf(buf, "rm -f %s/sessions/.default/views/*", datadir);
	system(buf);
	data = new Data();
	import_files.empty();
	import_lock = FALSE;

	action = 0;
	devicebit = positionbit = FALSE;
	strcpy(session_name, "");

	save_msg.setTitle("Save Session");
	save_msg.setLabel("Save current session?");
	save_msg.setButtons(1);
	gtk_signal_connect (GTK_OBJECT (save_msg.button_yes), "clicked",
		GTK_SIGNAL_FUNC (have_title_cb), this);
	gtk_signal_connect (GTK_OBJECT (save_msg.button_no), "clicked",
		GTK_SIGNAL_FUNC (action_cb), this);
	gtk_signal_connect (GTK_OBJECT (save_msg.button_cancel), "clicked",
		GTK_SIGNAL_FUNC (close_cb), save_msg.window);

	overwrite_msg.setTitle("Overwrite Session");
	overwrite_msg.setLabel("Session already exists.\nOverwrite session?");
	overwrite_msg.setButtons(2);
	gtk_signal_connect (GTK_OBJECT (overwrite_msg.button_yes), "clicked",
		GTK_SIGNAL_FUNC (save_session_cb), this);
	gtk_signal_connect (GTK_OBJECT (overwrite_msg.button_no), "clicked",
		GTK_SIGNAL_FUNC (close_cb), overwrite_msg.window);

	failure_msg.setTitle("Load Failed");
	failure_msg.setButtons(3);
	gtk_signal_connect (GTK_OBJECT (failure_msg.button_yes), "clicked",
		GTK_SIGNAL_FUNC (close_cb), failure_msg.window);

	load_session_list.setTitle("Load Session");
	gtk_signal_connect (GTK_OBJECT (load_session_list.entry), "activate",
		GTK_SIGNAL_FUNC(load_session_cb), this);
	gtk_signal_connect (GTK_OBJECT (load_session_list.button_accept), "clicked",
		GTK_SIGNAL_FUNC (load_session_cb), this);
	gtk_signal_connect (GTK_OBJECT (load_session_list.button_cancel), "clicked",
		GTK_SIGNAL_FUNC (close_cb), load_session_list.window);

	save_session_list.setTitle("Save Session");
	gtk_signal_connect (GTK_OBJECT (save_session_list.entry), "activate",
		GTK_SIGNAL_FUNC(unique_name_cb), this);
	gtk_signal_connect (GTK_OBJECT (save_session_list.button_accept), "clicked",
		GTK_SIGNAL_FUNC (unique_name_cb), this);
	gtk_signal_connect (GTK_OBJECT (save_session_list.button_cancel), "clicked",
		GTK_SIGNAL_FUNC (close_cb), save_session_list.window);

	progress.setTitle("Progress");
}

Session::~Session()
{
	if (data!=NULL) delete data;
	// FIXME cleanup import_files, filters, views
}

void Session::new_session(int type)
{
	GError *error;

	action = type;

	if ((devicebit==TRUE) || (positionbit==TRUE))
	{
		if ((main->gui->preferences->save_on_quit_flag==TRUE) && (action==4))
		{
			printf("Doing autosaveonquit\n");
			have_title();
		}else
			save_msg.show();
	}else
		session_action();
}

void Session::save()
{
	if ((devicebit==TRUE) || (positionbit==TRUE))
	{
		have_title();		
	}
}

void Session::have_title()
{
	save_msg.hide();
	if (strcmp(session_name, "")==0)
	{
		save_session_list.init(datadir);
		save_session_list.show();
	}else
		save_session();
}

void Session::session_action()
{
	struct stat statbuf;
	gchar buf[256];
	Link *ln;
	Filter *f;
	View *v;

	save_msg.hide();
	if (action==1)	// new
	{
		sprintf(buf, "rm -f %s/sessions/.default/views/*", datadir);
		system(buf);
		if (data!=NULL) delete data;
		data = new Data();
		devicebit = positionbit = FALSE;
		strcpy(session_name, "");
		main->setTitle();
		strcpy(data->startup_view, "");
		main->gui->update_views();
		main->gui->update_lists();
	}else
	if (action==2)	// open
	{
		load_session_list.init(datadir);
		load_session_list.show();
	}else
	if (action==3)	// saveas
	{
		// copy over views from previous session
		if (strcmp(orig_session_name, "")==0)
			strcpy(orig_session_name, ".default");
		sprintf(buf, "%s/sessions/%s", main->session->datadir, session_name);
		if (stat(buf, &statbuf) == -1) mkdir(buf, 488); // rwxr x
		sprintf(buf, "%s/sessions/%s/views", main->session->datadir, session_name);
		if (stat(buf, &statbuf) == -1) mkdir(buf, 488); // rwxr x
		sprintf(buf, "cp %s/sessions/%s/views/* %s/sessions/%s/views",
			main->session->datadir, orig_session_name,
			main->session->datadir, session_name);
		system(buf);
		main->setTitle();
	}else
	if (action==4)	// exit
	{
		main->quit();
	}else
		main->setTitle();
}

void Session::unique_name()
{
	char buf[256];
	SessionName *nm;
	Link *ln;
	int match;

	// extract name
	strcpy(buf, save_session_list.getName());
	if (strcmp(buf, "")==0) return;

	match = 0;
	for (ln=save_session_list.names.get_head(); ln!=NULL; ln=ln->next)
	{
		nm = (SessionName*) ln->item;
		if (strcmp(buf, nm->label)==0) match=1;
	}
	if ((match==1) && (strcmp(buf, session_name)!=0))
	{
		overwrite_msg.show();
	}else
	{
		strcpy(session_name, save_session_list.getName());
		save_session();
	}
}

void Session::save_session()
{
	struct stat statbuf;
	Filter *f;
	View *v;
	Link *ln;
	xmlDocPtr doc;
	char fname[256];
	GError *error;
	gboolean ch;

	// FIXME if fails during save call failed dialog
	overwrite_msg.hide();

	// make sure directory exists
	sprintf(fname, "%s/sessions/%s", datadir, session_name);
	if (stat(fname, &statbuf) == -1) mkdir(fname, 488); // rwxr x

	// save topology
//	if ((devicebit==TRUE) || (positionbit==TRUE) || (stat(fname, &statbuf)==-1))
	{
		// printf("SAVING NETWORK FILE\n");
		sprintf(fname, "%s/sessions/%s/network", datadir, session_name);
		data->savexml(fname);
		devicebit = FALSE;
	}

	session_action();
}

void Session::load_session(char *name)
{
	struct stat statbuf;
	xmlDocPtr doc, doc2;
	xmlNodePtr xmlnode;
	char buf[256];
	View *v;
	Link *ln;
	Data *d;
	gboolean forceflag;
	float t1, t2, sc;

	load_session_list.hide();

	if (strcmp(name, "")==0)
	{
		sprintf(buf, "Empty session name");
		failure_msg.setLabel(buf);
		failure_msg.show();
		return;
	}

//	sprintf(buf, "%s/sessions/%s/filters", datadir, name);
//	doc = xmlParseFile(buf);

	d = new Data();
	sprintf(buf, "%s/sessions/%s/network", datadir, name);
	d->loadxml(buf);
	// extra stuff to finish building the data structures
	//	main->topology->buildRestForLoad();
	// if we are just updating then copy over visual stuff
	// if we are loading a new diagram (accessing remotely) leave random OR
	//   if has a default view apply view

	// if everything ok so far delete existing structures and create new
	delete data;
	data = d;

	// here we apply a startup view if one is named
	if (strcmp(data->startup_view, "")!=0)
	{
		sprintf(buf, "%s/sessions/%s/views/%s", datadir, name, data->startup_view);
printf("LOADING STARTUP VIEW %s\n", buf);
		v = new View();
		v->load(buf);
		v->copy_to_topology(data);
		delete v;
		main->gui->update_views();
	}

	forceflag = main->gui->monitor->viewer->animate_version;
	main->gui->monitor->viewer->animate_version = FALSE;

	data->sizeshow();
	t1 = float((data->width)/(float)(main->gui->monitor->viewer->width-(BORDER*2)));
	t2 = float((data->height)/(float)(main->gui->monitor->viewer->height-(BORDER*2)));
	sc = (t1>t2)?t1:t2;
	if (sc<1.0)
		main->gui->monitor->viewer->scale = 1.0;
	else
		main->gui->monitor->viewer->scale = (float((int(sc*100))+1))/100;
	main->gui->monitor->viewer->xpos = (long) data->minx+(data->width/2);
	main->gui->monitor->viewer->ypos = (long) data->miny+(data->height/2);
	GTK_ADJUSTMENT(main->gui->monitor->viewer->hscale3_adj)->value = (float((int(sc*100))+1))/100;;
	main->gui->monitor->viewer->setScale(main->gui->monitor->viewer->scale);

	main->gui->monitor->viewer->animate_version = forceflag;

	strcpy(session_name, name);
	main->setTitle();
	main->gui->update_views();
	main->gui->update_lists();

	// ping stuff
	save_device_file();																// write out a devices file
	main->scheduler->addTask(session_name, 2, 0, 0);	// then assign a ping task
}

void Session::save_device_file()
{
	struct stat statbuf;
	FILE *f;
	char buf[256];
	Link *ln;
	Device *d;

	if (strcmp(session_name, "")!=0)
	{
		sprintf(buf, "%s/sessions/%s", datadir, session_name);
		if (stat(buf, &statbuf) == -1) mkdir(buf, 488); // rwxr x
		sprintf(buf, "%s/sessions/%s/devices", datadir, session_name);
	}else
	{
		sprintf(buf, "%s/sessions/.default", datadir);
		if (stat(buf, &statbuf) == -1) mkdir(buf, 488); // rwxr x
		sprintf(buf, "%s/sessions/.default/devices", datadir);
	}
	f = fopen(buf, "w");
	if (f)
	{
		for (ln=data->devices.get_head(); ln!=NULL; ln=ln->next)
		{
			d = (Device*) ln->item;
//			if (d->type!='u')
			if (!d->pingname.zero())
			{
				fprintf(f, "%c,%s,%s\n",
					d->grab_stats, d->pingname.getstring(), d->community);
			}
		}
		fclose(f);
	}
}

/*
int Session::uniquefiltername(Filter *orig, char *name)
{
	Link *ln;
	Filter *f;

	for (ln=filters.get_head(); ln!=NULL; ln=ln->next)
	{
		f = (Filter*) ln->item;
		if ((orig!=f) && (strcmp(name, f->name)==0))
			return 0;
	}
	return 1;
}

int Session::uniqueviewname(char *name)
{
	Link *ln;
	View *v;

	for (ln=views.get_head(); ln!=NULL; ln=ln->next)
	{
		v = (View*) ln->item;
		if (strcmp(name, v->name)==0)
			return 0;
	}
	return 1;
}
*/

void Session::import_init(gchar *fname)
{
	import_files.tag(g_strdup(fname));
}

void Session::import_kill()
{
	// FIXME stop current import and remove rest from list
	// connect up to cancel and destroy in progress window
}

void Session::import(gboolean show_progress)
{
	Link *ln;
	gchar line[256];
	gchar *str;

	if (import_lock==TRUE) return;
	import_lock = TRUE;
	if (show_progress==TRUE) progress.show();
	while (import_files.size()>0)
	{
		// import next file from list
		ln = import_files.get_head();
		str = (gchar*) ln->item;
		sprintf(line, "Import file\n%s", str);
		if (show_progress==TRUE) progress.setLabel(line);
		if (show_progress==TRUE) progress.setValue(0);
		import_file(str, show_progress);
		import_files.del(str);
		g_free(str);
	}
	if (show_progress==TRUE) progress.hide();
	import_lock = FALSE;
	devicebit = TRUE;
}

void Session::import_file(gchar *fname, gboolean show_progress)
{
	FILE *f;
	gchar line[256];
	gint count, num_lines;
	gchar description[256];
	gchar ipaddr[32];
	char hostname[256];
	char mac[256];
	char location[256];
	char contact[256];
	gint i, mask;
	IPAddress tmpip;
	MACAddress tmpmac;
	Link *ln;
	Subnet *s;
	Machine *m;

	// work out number of lines in file
	num_lines = linesInFile(fname);
	if (num_lines==0) return;

	count = 0;
	f = fopen(fname, "r");
	if (f)
	{
		while (fscanf(f, "%[^\n]\n", line)==1)
		{
			if (sscanf(line, "S \"%[^\"]\",%[^,],%d", description, ipaddr, &mask)==3)
			{
				tmpip.setstring(ipaddr);
				for (ln=data->subnets.get_head(); ln!=NULL; ln=ln->next)
				{
					s = (Subnet*) ln->item;
					if ((tmpip.compare(s->ip)==0) && (mask==s->mask))
					{
						strcpy(s->name, description);
					}
				}
			}else
			if (sscanf(line, "M \"%[^\"]\",%[^,],\"%[^\"]\",\"%[^\"]\",\"%[^\"]\",\"%[^\"]\"", hostname, ipaddr, mac, contact, location, description)==6)
			{
				tmpip.setstring(ipaddr);
				tmpmac.setstring(mac);
				// if mac given OR ip given
				if ((tmpmac.zero()==0) || (!tmpip.zero()))
				{
					for (ln=data->machines.get_head(); ln!=NULL; ln=ln->next)
					{
						m = (Machine*) ln->item;
						// if mac OR ip match
						if ((m->mac.compare(tmpmac)==1) || (m->ip.compare(tmpip)==0))
						{
							strcpy(m->name, hostname);
							m->mac.set(tmpmac);
							m->ipinv.set(tmpip);
							strcpy(m->location, location);
							strcpy(m->descr, description);
							strcpy(m->contact, contact);
						}
					}
				}
			}
			count++;
			if ((count%10)==0)
			{
				if (show_progress==TRUE)
					progress.setValue(((gfloat)(count))/(num_lines));
				while(gtk_events_pending()) gtk_main_iteration();
			}
		}
		fclose(f);
	}
}

/*

//	xmlDocPtr doc;
//	xmlNodePtr xmlnode;
	FILE *f;
	char line[1024];
	char hostname[256];
	char mac[256];
	char location[256];
	char description[256];
	char contact[256];
	char ipaddr[32];
	IPAddress tmpip;
	Link *ln;
	Machine *m;
	Progress prg;
	int i, count, mask, len;

	prg.show();
	sprintf(line, "Import file\n%s", fname);
	prg.setLabel(line);
	prg.setValue(0);

	count = 0;
	f = fopen(fname, "r");
	if (f)
	{
		if (((line[0]=='S') || (line[0]=='M')) && (line[1]==' '))
		{
			// do short version

			while (fscanf(f, "%[^\n]\n", line)==1)
			{
				if (line[0]=='S')
				{
					if (sscanf(line, "S \"%[^\"]\",%[^,],%d", description, ipaddr, &mask)==3)
					{
						//	printf("SHORT SUBNET %s\n", description);
						tmpip.setstring(ipaddr);
					}
				}else
				if (line[0]=='M')
				{
					if (sscanf(line, "M \"%[^\"]\",%[^,],\"%[^\"]\",\"%[^\"]\",\"%[^\"]\",\"%[^\"]\"", hostname, ipaddr, mac, contact, location, description)==6)
					{
						//	printf("SHORT MACHINE %s\n", hostname);
						tmpip.setstring(ipaddr);
						for (ln=data->machines.get_head(); ln!=NULL; ln=ln->next)
						{
							m = (Machine*) ln->item;
//							if (m->ip.compare(tmpip)==0)
//							{
//								i = strlen(hostname);
//								hostname[i-1] = '\0';
//								strcpy(m->name, hostname);
//								printf("MATCH %s %s\n", hostname, ipaddr);
//							}
						}
					}
				}

				count++;
				if ((count%10)==0)
				{
					prg.setValue(((gfloat)(count))/(linecount));
					while(gtk_events_pending()) gtk_main_iteration();
				}
			}
		}else
		{
			// do dns host
			
			while (fscanf(f, "%[^\n]\n", line)==1)
			{
				if (sscanf(line, "%s has address %s", hostname, ipaddr)==2)
				{
					tmpip.setstring(ipaddr);
					for (ln=data->machines.get_head(); ln!=NULL; ln=ln->next)
					{
						m = (Machine*) ln->item;
						if (m->ip.compare(tmpip)==0)
						{
							i = strlen(hostname);
							hostname[i-1] = '\0';
							strcpy(m->name, hostname);
						}
					}
				}
				count++;
				if ((count%10)==0)
				{
					prg.setValue(((gfloat)(count))/(linecount));
					while(gtk_events_pending()) gtk_main_iteration();
				}
			}
		}
		fclose(f);
	}
	prg.hide();
}

*/

