/* main.cc
 * Copyright (C) 2001,2002 The University of Newcastle upon Tyne
 *
 * Written by Paul Coates (Paul.Coates@ncl.ac.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <gnome.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <pwd.h>
#include "main.h"
#include "misc.h"
#include "header.h"

int main(int argc, char *argv[])
{
	Main m(argc, argv);
	return 0;
}

Main::Main(int argc, char *argv[])
{
	gchar *teststr;

	gnome_init("nomad-stats", VERSION, argc, argv);
	gconf_init (argc, argv, NULL);
	client = gconf_client_get_default();
  gconf_client_add_dir(client, "/apps/nomad",
		GCONF_CLIENT_PRELOAD_NONE, NULL);

	/* check if gconf schemas are working */
	teststr = gconf_client_get_string (client, "/apps/nomad/gconf_test", NULL);
	if (!teststr)
	{
		printf("Can't find a schema for nomad preferences. Check your gconf setup\n");
		return;
	}
	else
	{
		g_free (teststr);
	}

	if (argc!=2)
	{
		printf("usage:%s <session name>\n", argv[0]);
		return;
	}

	// FIXME get num processes from command line if want different from default

	strcpy(session, argv[1]);

	run();
}

Main::~Main()
{
	quit();
}

// GConf - too dangerous to use NOTIFY, instead we wait until the end of the current run, grab the new settings, then start with new settings

void Main::run()
{
	char buf[256];
	struct passwd *pwd;
	Process *p;
	Link *ln;
	char *md;
	Task *t;
	int i, flag;
printf("Main::run\n");
	md = getenv("NOMAD_DIR");
	if (md!=NULL) { strcpy(datadir, md); } else
	{ pwd = getpwuid(getuid());
		sprintf(datadir, "%s/%s", pwd->pw_dir, DATA_DIR);
	}

	for (i=1; i<=NUM_PROCESSES; i++)
	{
		p = new Process();
		p->create();
		processes.tag(p);
	}

	stat_tm = time(NULL);
	stat_tasks.empty();

	// the following works as the task list starts empty so the first thing that
	// happens is create a new task list, then send keep alive messages which
	// force loading defaults from gconf

printf("Main::run starting main while loop\n");
	while(1)
	{
		// check for processes ending and set appropriate flags
		for (ln=processes.get_head(); ln!=NULL; ln=ln->next)
		{
			p = (Process*) ln->item;
			if (p->get(buf))
			{
				check_incoming_task(&stat_tasks, p);
				p->tid = 0;
			}
		}

		process_tasks(&stat_tasks);

		flag = 0;
		for (ln=stat_tasks.get_head(); ln!=NULL; ln=ln->next)
		{
			t = (Task*) ln->item;
			if (t->status == 0) flag = 1;
		}
		if (flag == 0)	// all processes returned
		{
			stat_tm = time(NULL);
			tasks_from_devices(&stat_tasks, "nomad-traffic");

			// reset keepalive timeouts, now is also a good time to
			// regrab gconf values to get up-to-date details
			for (ln=processes.get_head(); ln!=NULL; ln=ln->next)
			{
				p = (Process*) ln->item;
				if (p->tid==0)
				{
					printf("Sending gconf to process %d\n", p->pid);
					p->tid = 1;
					p->send("gconf");
				}
			}
		}
//		status();
		usleep(2000000);
	}
}

// check the task list for the task that is being run by the returning process
void Main::check_incoming_task(List *lst, Process *p)
{
	Link *ln2;
	Task *t;

	for (ln2=lst->get_head(); ln2!=NULL; ln2=ln2->next)
	{
		t = (Task*) ln2->item;
		if (t->pid == p->pid)
		{
			printf("Process %lu returned a task\n", p->pid);
			t->pid = 0;
			t->status = 1;
		}
	}
}

// create a task list from the devices file
void Main::tasks_from_devices(List *lst, char *str)
{
	char buf[256];
	char buf2[256];
	char buf3[256];
	char buf4[256];
	char line[256];
	FILE *f;
	Process *p;
	Task *t;
	Link *ln;
	int i, count;
	gchar *str2;
	gint val;

	// delete any existing tasks
	for (ln=lst->get_head(); ln!=NULL; ln=ln->next)
	{ t = (Task*) ln->item; delete t; } lst->empty();

	// create a new list

//	printf("CREATE A NEW TASK LIST\n");
	count = 0;
	sprintf(buf, "%s/sessions/%s/devices", datadir, session);
	f = fopen(buf, "r");
	if (f)
	{
		while(fgets(line, 256, f) != NULL)
		{
			if (sscanf(line, "%[^,],%[^,],%[^\n]\n", buf2, buf3, buf4)==3)
			{
				t = new Task();
				sprintf(t->cmnd, "%s %s %s", str, buf3, buf4);
				lst->tag(t);
				count++;
//printf("%s\n", t->cmnd);
			}
		}
		fclose(f);
	}

//printf("COUNT = %d tasks/devices\n", count);

	str2 = gconf_client_get_string(client, CONFIG_SAMPLE_RATE, NULL);
	val = timestr_to_int(str2);
//	printf("%d\n", val);
	// minimise time to 2 minutes
	if (val<MIN_STATS_TIME) val = MIN_STATS_TIME;
	i = 0;
	for (ln=lst->get_head(); ln!=NULL; ln=ln->next)
	{
		t = (Task*)ln->item;
		t->ts = (int)((val*(i+1))/count);
//printf("Task %d time offset is %d\n", i, t->ts);
		i++;
	}

	for (ln=processes.get_head(); ln!=NULL; ln=ln->next)
	{
		p = (Process*)ln->item;
		p->frequency = val;
//		printf("process status pid=%d tid=%d\n", p->pid, p->tid);
	}
}

// take in a task list and assign what you can to any free processes
void Main::process_tasks(List *lst)
{
	Link *ln, *ln2;
	Process *p;
	Task *t;
	time_t now_tm = time(NULL);

//	printf("In process tasks (i.e. send to child processes)\n");
	for (ln=processes.get_head(); ln!=NULL; ln=ln->next)
	{
		p = (Process*)ln->item;
		for (ln2=lst->get_head(); ln2!=NULL; ln2=ln2->next)
		{
			t = (Task*) ln2->item;
			if ((t->status==0) && (t->pid==0) && (p->tid==0) &&
					(now_tm >= stat_tm+t->ts))
			{
				printf("Sending task %s to process %d\n", t->cmnd, p->pid);
				t->pid = p->pid;
				p->tid = 1;
				strcpy(p->cmnd, t->cmnd);
				p->send(t->cmnd);
			}
		}
		// if process not needed send a keep alive message
		if (p->tid == 0)
		{
//			printf("Sending keepalive to process %d\n", p->pid);
			p->tid = 1;
			p->send("keepalive");
		}
	}
}

void Main::status()
{
	Link *ln;
	Process *p;
	Task *t;
	int i;

	i=1;
	for (ln=processes.get_head(); ln!=NULL; ln=ln->next)
	{
		p = (Process*)ln->item;
		printf("STATUS - Process %2d %6d %6d %4d %s\n", i, p->pid, p->tid, p->frequency, p->cmnd);
//		if (p->tid!=0)
//		{
//			printf("%s", p->cmnd);
//		}
//		printf("\n");
		i++;
	}


	for (ln=stat_tasks.get_head(); ln!=NULL; ln=ln->next)
	{
		t = (Task*)ln->item;
		printf("STATUS - Task %4d %d %8d %s\n", t->pid, t->status, t->ts, t->cmnd);
	}
}

void Main::quit()
{
	Link *ln;

	for (ln=processes.get_head(); ln!=NULL; ln=ln->next)
		((Process*)ln->item)->send("quit");
}

